
#include <fstream.h>
#include <time.h>
#include "boards.h"

#define MAX_MOVES                       12500
#undef MEMORY_INTENSIVE

#ifdef MEMORY_INTENSIVE

struct AnaObject {
    bool is_word;
    int name;
    struct AnaObject *a[26];
};

#else

struct AnaObject {
    int info;                   // Bit 0 (LSB): is this a word?
                                // Bit 1..26  : set if ('a'+bit-1) is linked
    int name;
    struct AnaObject **a;       // variably-sized array of AnaObjects
};

#endif

struct Move {
    int which_word;
    int column;
    int row;
    int blank[2];
    bool across;
    bool is_bingo;
    int triple;
    int score;
    int equity;
};

struct Fish {
    bool is_fish;
    int fish_tile;
    int rid;
};

enum invalid_moves {center_square, one_line, tiles_touch, invalid_word,
                    no_tiles};

struct UserMove {
       bool valid_move;
       int scores;
       enum invalid_moves reason;
};

class Dictionary {
public:
    Dictionary(int);
    ~Dictionary();
    bool load_dictionary(char *);
    int check_word(char *);
    char *report_word(int);
    void expurgate_dictionary(bool);
private:
    void add_word(char *);
    void remove_word(char *);
    char **word;
    int nwords;
};

class Well {
public:
    Well();
    char draw_from_well();
    bool is_well_empty();
    bool remove_from_well(int);
    void return_to_well(char);
    void refill_well();
    int t_left() { return tiles_left; }
    float prob(int);
private:
    char contents[100];
    int tiles_left;
};

class Rack {
public:
    Rack();
    void add_letter(char);
    bool del_letter(char);
    void remove_letter(int);
    void empty_rack(Well *);
    void shuffle();
    void fill_from_well(Well *);
    void print_rack(ostream &);
    void draw_rack();
    void inc(char);
    int exchange_tiles(Well *);
    int n_tiles() { return ntiles; }
    int how_many(int i) { return c[i]; }
    int return_letter(int);
    void dump_rack();
    void copy_rack(int *);
    char *exchange_string(char *, Well *);
    char r_display(int i) { return display[i]; }
    void s_display(int i, char ch) { display[i]=ch; }
private:
    int ntiles;
    int c[27];
    char display[7];
};

class SBoard {
public:
    SBoard(int);                /* the argument is which board to use */
    void find_moves(Rack *, Dictionary *, Well *);
    void find_fish_moves(Rack *, Dictionary *, Well *);
    void redraw(bool, bool);
    void move_set(int, Dictionary *, Rack *);
    void move_set_imaginary(int, Dictionary *, Rack *);
    int user_move(Rack *, Dictionary *, Well *, struct PlayerDesc *, struct Options *, ofstream &);
    int computer_move(Rack *, Dictionary *, Well *, struct PlayerDesc *, struct Options *, ofstream &, bool, int, int);
    struct Move *report_move(int);
    void print_move(struct Move *, int, Dictionary *, ostream &);
    int n_moves() { return nmoves; }
    char board(int row, int col) { return contents[row][col]; }
    bool blank(int row, int col) { return is_blank[row][col]; }
    void clear_board(Well *);
    void set_on_board(int, int, char, int);
    void dump_board();
    void print_board(ostream &);
    void set_flag_d(bool b) { play_defensively=b; }
    bool add_comment_information;
    bool two_player_end;
    Rack *ogame;
private:
    void find_anagrams(Rack *, int, int, int);
    void valid_anagrams(struct AnaObject *, bool *, Rack *, int, int, int, int, int, int, int);
    void perform_first_move(Rack *, Dictionary *, Well *);
    void correct_moves(Rack *, Dictionary *, Well *);
    void defensive_strategy(Rack *, Dictionary *);
    void set_imaginary(int, Dictionary *);
    void clear_imaginary();
    void check_exchange(Rack *, Dictionary *);
    void end_game(Rack *, Dictionary *);
    void two_player(Rack *, Dictionary *, Rack *, Well *);
    void danger_spots(int);
    void init_open();
    void print_user_score(Dictionary *, struct PlayerDesc *);
    int score_move(int, Dictionary *);
    int compute_equity(int, Rack *, Dictionary *, Well *);
    int with_rack(Rack *, Dictionary *, int);
    int rid_of(Rack *, Dictionary *, int, struct Fish *);
    int tree_leaf(int *, Dictionary *, int);
    int nmoves;
    int still_out[27];
    int my_score;
    int their_score;
    struct Move *moves;
    struct Move *their_move;
    struct Move users_move;
    struct UserMove *check_user_move(Dictionary *, struct PlayerDesc *, ofstream &, bool);
    Rack *orack;
    char contents[15][15];
    bool move_uses(int, int, Dictionary *);
    bool remove_letters(int *, int, Dictionary *);
    bool legal_letters[15][15][26];
    bool must_include[15][15];
    bool is_blank[15][15];
    bool is_imaginary[15][15];
    bool shape[15][15];
    bool open[4];
    bool across;
    bool redirect;
    int bonuses[15][15];
    int arrowx;
    int arrowy;
    bool arrow_across;
    bool is_first_move;
    bool play_defensively;
    bool kill_blank_equity;
    bool generate_commentary_info;
};

class Commentary {
public:
    Commentary(char *);
    void generate_commentary(int, Rack *, Dictionary *, Well *, SBoard *, struct PlayerDesc *, struct Moves *);
    void output_computer_move(struct PlayerDesc *, struct Move *);
    void init_commentary(ofstream &);
    struct Move move_analyze;
    struct {
           char rack_leave[7];
           bool front_hooks[26];
           bool front_double_word;
           bool front_triple_word;
           bool back_hooks[26];
           bool back_double_word;
           bool back_triple_word;
           bool keeps_U;
           bool saves_X;
           int combination3;
           int vowels;
           int consonants;
           int tl_opened;
           int dl_opened;
    } move_information[MAX_MOVES];
private:
    struct {
       int bingoes;
       int rack_leave_rating;
       int number_moves;
       int score;
       bool opened_triple_last;
       bool closed_triple_last;
       bool exchanged_last;
       bool blocked_opponent;
       bool played_U;
    } analysis[4];
    ofstream fout;
};

struct PlayerDesc {
    char *name;
    char *last_move;
    bool is_human;
    bool playing;
    int difficulty;             /* only used when is_human==false */
    int score;
    int last_score;
};

struct Options {
    bool expurgate;
    bool timers;
    bool hints;
    bool flash_on_move;
    bool another_game;
    bool recap;
    int shape;
};

class Game {
public:
     Game();
     ~Game();
     void assign_players(struct PlayerDesc *);
     void assign_dictionary(Dictionary *);
     void assign_options(struct Options *);
     void play_game();
private:
     void player_status(int);
     int nplayers;
     SBoard *board;
     Well well;
     struct PlayerDesc *players;
     struct Options *game_options;
     Dictionary *master;
     Rack *racks;
};

const int letter_val[26]={1,3,3,2,1,4,2,4,1,8,5,1,3,1,1,3,10,1,1,1,1,4,4,8,4,10};

const int letter_dist[27]={9,2,2,4,12,2,3,2,9,1,1,4,2,6,8,2,1,6,4,6,4,2,2,1,2,1,2};


